

#include <SDL2/SDL.h>

#include <chrono>
#include <format>
#include <memory>
#include <print>
#include <random>
#include <source_location>
#include <stdexcept>

namespace {

auto sdl_throw_current_error(const std::source_location location) {
	throw std::runtime_error{std::format(
		"error at {}:{}:{} {}", location.file_name(), location.line(),
		location.column(), SDL_GetError()
	)};
}

auto sdl_check(
	int errc,
	const std::source_location location = std::source_location::current()
) {
	if (errc != 0) { sdl_throw_current_error(location); }
}

// make sure someone already owns passed sdl_obj, or memory will be leaked
auto sdl_check(
	auto* sdl_obj,
	const std::source_location location = std::source_location::current()
) {
	if (sdl_obj == nullptr) { sdl_throw_current_error(location); }
	return sdl_obj;
}

template <typename SDLT, auto deletef>
struct sdl_wrapper_t : std::unique_ptr<SDLT, decltype(deletef)> {
	explicit sdl_wrapper_t(SDLT* p)
		: std::unique_ptr<SDLT, decltype(deletef)>{p, deletef} {
		sdl_check(p);
	}

	operator SDLT*() { return std::unique_ptr<SDLT, decltype(deletef)>::get(); }
};

using window_t = sdl_wrapper_t<SDL_Window, SDL_DestroyWindow>;
using surface_t = sdl_wrapper_t<SDL_Surface, SDL_FreeSurface>;
using renderer_t = sdl_wrapper_t<SDL_Renderer, SDL_DestroyRenderer>;

auto operator+(SDL_Point a, SDL_Point b) {
	return SDL_Point{.x = a.x + b.x, .y = a.y + b.y};
}

auto operator<=>(const SDL_Point& a, const SDL_Point& b) {
	auto y_comp = a.x <=> b.x;
	if (y_comp != std::strong_ordering::equal) { return y_comp; }
	return a.y <=> b.y;
}

auto operator==(const SDL_Point& a, const SDL_Point& b) {
	return a.x == b.x && a.y == b.y;
}

// constexpr int width = 1000;
// constexpr int height = 1000;
constexpr int width = 3440;
constexpr int height = 1440;

template <auto EF>
struct [[nodiscard("give this a name so SDL_Quit is called at the end"
)]] Defer {
	Defer() = default;
	Defer(const Defer&) = delete;
	Defer(Defer&&) = delete;
	auto operator=(Defer&&) = delete;
	auto operator=(const Defer&) = delete;

	~Defer() { EF(); }
};

using clock = std::chrono::steady_clock;

auto poll_events(bool& continu) {
	for (SDL_Event e; SDL_PollEvent(&e);) {
		if (e.type == SDL_QUIT) { continu = false; }
	}
}
} // namespace

int main() try {
	sdl_check(SDL_Init(SDL_INIT_VIDEO));
	Defer<SDL_Quit> defer_SDL_Quit;

	window_t window{SDL_CreateWindow(
		"random walk", SDL_WINDOWPOS_UNDEFINED, SDL_WINDOWPOS_UNDEFINED, width,
		height, 0 // SDL_WINDOW_FULLSCREEN
	)};
	renderer_t renderer{SDL_CreateRenderer(window, -1, SDL_RENDERER_ACCELERATED)
	};
	SDL_SetHint(SDL_HINT_RENDER_SCALE_QUALITY, "1");
	sdl_check(SDL_SetRenderDrawColor(renderer, 0, 0, 0, SDL_ALPHA_OPAQUE));
	sdl_check(SDL_RenderClear(renderer));
	sdl_check(SDL_SetRenderDrawColor(renderer, 0, 0, 255, SDL_ALPHA_OPAQUE));
	{
		// dump renderer info
		SDL_RendererInfo renderer_info;
		SDL_GetRendererInfo(renderer, &renderer_info);
		std::println("Renderer name: {}", renderer_info.name);
		std::println("Texture formats: ");
		for (Uint32 i = 0; i < renderer_info.num_texture_formats; i++) {
			std::puts(SDL_GetPixelFormatName(renderer_info.texture_formats[i]));
		}
	}

	// no clue who owns this
	auto* texture = SDL_CreateTexture(
		renderer, SDL_PIXELFORMAT_ARGB8888, SDL_TEXTUREACCESS_TARGET, width,
		height
	);

		SDL_Point pos{.x = width / 2, .y = height / 2};
	constexpr SDL_Point directions[] = {
		{.x = -1, .y = 0}, {.x = 1, .y = 0}, {.x = 0, .y = -1}, {.x = 0, .y = 1}
	};

	std::mt19937_64 rne(std::random_device{}());
	std::uniform_int_distribution dist(0, 3);
	//                              aarrggbb??
	constexpr Uint32 bg_color = 0xffdbd7c0;
	constexpr Uint32 walk_color = 0xff0d0f14;
	constexpr auto batchsize = 1;

	const auto start_time = clock::now();
	auto next_poll_events_time = start_time;
	auto frame_start = start_time;
	clock::duration frame_time;
	for (bool continu = true; continu;) {
		// measure frame_time
		const auto now = clock::now();
		frame_time = now - frame_start;
		frame_start = now;

		if (now >= next_poll_events_time) {
			next_poll_events_time = now + std::chrono::milliseconds{200};
			poll_events(continu);
			std::println(
				"last frame_time={} batchsize={}", frame_time, batchsize
			);
		}

		sdl_check(SDL_SetRenderTarget(renderer, texture));
		for (auto step = 0z; step < batchsize; ++step) {
			SDL_Point newpoint;
			do {
				newpoint = pos + directions[dist(rne)];
			} while (newpoint.x < 0 or newpoint.x >= width or newpoint.y < 0 or
			         newpoint.y >= height);
			pos = newpoint;
			sdl_check(SDL_RenderDrawPoint(renderer, pos.x, pos.y));
		}

		sdl_check(SDL_SetRenderTarget(renderer, nullptr));
		sdl_check(SDL_RenderCopy(renderer, texture, nullptr, nullptr));
		SDL_RenderPresent(renderer);
	}

	// leaking memory, nothing I can do
} catch (const std::exception& e) {

	std::puts(e.what());
	return EXIT_FAILURE;
}